import os
from pathlib import Path
from typing import List

from tree_sitter import Node, Language, Parser, Tree

from tree_hugger.exceptions import ParserLibraryNotFoundError, SourceFileNotFoundError, QueryFileNotFoundError
from tree_hugger.core.queries import Query


def match_from_span(node: Node, lines: List) -> str:
    line_start = node.start_point[0]
    line_end = node.end_point[0]
    char_start = node.start_point[1]
    char_end = node.end_point[1]
    if line_start != line_end:
        return '\n'.join([lines[line_start][char_start:]] + lines[line_start+1:line_end] + 
                          [lines[line_end][:char_end]])
    else:
        return lines[line_start][char_start:char_end]


class BaseParser(object):
    """
    Base parser exposes the common interface that we extend per language
    """
    
    DEFAULT_QUERY_FILE_PATH = str(Path("tree_hugger") / "assets" / "queries" / "queries.yml")

    def __init__(self, language: str, query_class_name: str, library_loc: str=None, query_file_path: str=None):
        if os.getenv("TS_LIB_PATH") is not None and library_loc is None:
            library_loc = os.getenv("TS_LIB_PATH")
        
        if not library_loc:
            raise ParserLibraryNotFoundError("Parser library path is 'None'. Please either set up the environment or call the constructor with the path")

        if not Path(library_loc).exists() or not Path(library_loc).is_file():
            raise ParserLibraryNotFoundError(f"Parser library '{library_loc}' not found. Did you set up the environement variables?")
        
        if os.getenv("QUERY_FILE_PATH") is not None and query_file_path is None:
            query_file_path =  os.getenv("QUERY_FILE_PATH")
        
        if not query_file_path:
            query_file_path = BaseParser.DEFAULT_QUERY_FILE_PATH

        
        self.language = Language(library_loc, language)
        self.parser = Parser()
        self.parser.set_language(self.language)
        self.qclass = Query(query_file_path)
        self.QUERIES = self.qclass[query_class_name]
    
    def _run_query_and_get_captures(self, q_name: str, root_node:  Node) -> List:
        """
        Runs a query on the the language and returns the raw spans.

        Sometimes you may need to do to some specilized processing on the returned lines
        Check languane specific codes to see what that can be.
        """
        query = self.language.query(self.QUERIES[q_name])
        captures = query.captures(root_node)
        return captures
    
    def parse_file(self, file_path: str):
        """
        Parses a single file and retunrs True if success
        """
        if not Path(file_path).exists() or not Path(file_path).is_file():
            raise SourceFileNotFoundError(f"Source file {file_path} not found")
        try:
            with open(file_path, encoding='utf-8') as f:
                blob = f.read()
        except UnicodeDecodeError:
            return False
        self.raw_code = blob
        self.splitted_code = blob.split("\n")
        self.tree :Tree = self.parser.parse(bytes(blob.encode('utf-8')))
        self.root_node :Node = self.tree.root_node
        return True
    
    def sexp(self):
        return self.tree.root_node.sexp()
    
    def reload_queries(self, query_file_path: str=None):
        """
        Reloads the query file and the internal data structure.

        This is a temporary method and it should not be a part of the API. 
        mainly to dynamically reload the queries while in development for 
        IPython console
        """
        if os.getenv("QUERY_FILE_PATH") is not None and query_file_path is None:
            query_file_path =  os.getenv("QUERY_FILE_PATH")
        
        self.qclass = Query(query_file_path)
        self.QUERIES = self.qclass['python_queries']
    
    def _has_children(self, node: Node) -> bool:
        return len(node.children) > 0



import os
import logging
from pathlib import Path
from collections import UserDict

import yaml

import tree_hugger.setup_logging
from tree_hugger.exceptions import QueryFileNotFoundError


class Query(UserDict):

    data = {}

    def __init__(self, query_file_path: str):
        if not Path(query_file_path).exists() or not Path(query_file_path).is_file():
            raise QueryFileNotFoundError(f"Cound not find {query_file_path}")
        
        with open(query_file_path) as f:
            self.update(yaml.load(f, Loader=yaml.FullLoader))



import pathlib
from setuptools import find_packages, setup

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()


setup(
name='tree-hugger',
description="A light-weight, Extendable, high level, universal code parser built on top of tree-sitter",
long_description=README,
long_description_content_type="text/markdown",
url="https://github.com/autosoft-dev/tree-hugger",
author="CodistAI",
author_email="shubhadeep@cdist-ai.com",
include_package_data=True,
license="MIT",
classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
],
version="0.2.2",
packages=find_packages(exclude=("tests",)),
install_requires=["tree-sitter", "pygit2", "pytest", "PyYAML"],
entry_points = {
    'console_scripts': ['create_libs=tree_hugger.cli.create_libs:main'],
},
)



from argparse import ArgumentParser
import logging
import os
from pathlib import Path
import shutil

import pygit2
from tree_sitter import Language

import tree_hugger.setup_logging


REPO_PREFIX = 'https://github.com/tree-sitter/tree-sitter-'
TEMP_DOWNLOAD_PATH = "/tmp/tree-sitter-repos"

parser = ArgumentParser()
parser.add_argument("langs",
                    type=str,
                    nargs="+",
                    help="Give the name of languages for tree-sitter (php, python, go ...)")
parser.add_argument("-c",
                    "--copy-to-workspace",
                    action='store_true',
                    help="Shall we copy the created libs to the present dir? (default: False)")
parser.add_argument("-l",
                    "--lib-name",
                    type=str,
                    default="my-languages.so",
                    required=False,
                    help="The name of the generated .so file")



repo_path = lambda x:  f"{TEMP_DOWNLOAD_PATH}/{x}"
lib_path = lambda x: os.getcwd() if x.copy_to_workspace else TEMP_DOWNLOAD_PATH


def clone_repo(lang_name):
    """
    Clone the necessary repo
    """
    if Path(repo_path(lang_name)).exists():
        shutil.rmtree(Path(repo_path(lang_name)))
    repo_url = f"{REPO_PREFIX}{lang_name}"
    try:
        r = pygit2.clone_repository(repo_url, repo_path(lang_name))
        return True if r else False
    except pygit2.GitError:
        return False


def make_tree_sitter_lib(args, lang_repo_list):
    """
    Create the library from the repos
    """
    lp = lib_path(args)
    lib_name = args.lib_name
    full_lib_creation_path = f"{lp}/{lib_name}"
    
    if Path(full_lib_creation_path).exists():
        os.remove(full_lib_creation_path)
    
    return Language.build_library(full_lib_creation_path, lang_repo_list)


def main():
    args = parser.parse_args()
    repo_arr = []
    for lang_name in args.langs:
        logging.info(f"Cloneing {lang_name} repo from tree-sitter collections")
        ret = clone_repo(lang_name)
        if ret:
            logging.info("Clone success")
            repo_arr.append(repo_path(lang_name))
        else:
            logging.error("Clone falied. Are you sure the language binding exists?")

    if repo_arr:
        logging.info(f"Creating the library {args.lib_name} at {lib_path(args)}")
        ret = make_tree_sitter_lib(args, repo_arr)
        if ret:
            logging.info("Finished creating library!")
    else:
        logging.info("No libraries could be created. Please try again")



import re
from typing import List, Dict
import logging

from tree_sitter import Tree, Node

from tree_hugger.core.code_parser import BaseParser, match_from_span
import tree_hugger.setup_logging

TRIPPLE_QUOTE = '"""'
TRIPPLE_SINGLE_QUOTE = "'''"
TRIPPLE_QUOTE_NUMPY_STYLE = 'r"""'
TRIPPLE_SINGLE_QUOTE_NUMPY_STYLE = "r'''"

starts_with_tripple_quote = lambda x: x.startswith(TRIPPLE_QUOTE)
starts_with_tripple_single_quote = lambda x: x.startswith(TRIPPLE_SINGLE_QUOTE)
starts_with_numpy_style_tripple_quote = lambda x: x.startswith(TRIPPLE_QUOTE_NUMPY_STYLE)
starts_with_numpy_style_tripple_single_quote = lambda x: x.startswith(TRIPPLE_SINGLE_QUOTE_NUMPY_STYLE)

regex = r"([ ]{2,})"


class PythonParser(BaseParser):
    """
    The PythonParser class, extending the BaseParser and supplying some easy-to-use API s for mining code files

    @TODO - Implement a tree walker.
    """

    def __init__(self, library_loc: str=None, query_file_path: str=None):
        super(PythonParser, self).__init__('python', 'python_quaries', library_loc, query_file_path)
    
    def _strip_py_doc_string(self, dt: str, strip_quotes: bool) -> str:
        try:
            if starts_with_tripple_quote(dt):
                regex = r"\"{3}[\s\S]*?\"{3}"
            elif starts_with_tripple_single_quote(dt):
                regex = r"\'{3}[\s\S]*?\'{3}"
            elif starts_with_numpy_style_tripple_quote(dt):  # For Spinhx (numpy) style docstring with \ escapes
                # Check this - https://stackoverflow.com/questions/46543194/does-r-stand-for-something-in-sphinx
                regex = r"r\"{3}[\s\S]*?\"{3}"
            elif starts_with_numpy_style_tripple_single_quote(dt):  # For Spinhx (numpy) style docstring with \ escapes
                regex = r"r\'{3}[\s\S]*?\'{3}"
            if regex is None:
                logging.info(f"not a docstring {dt}")
            matches = re.search(regex, dt)
            return_dt = matches.group()
            if not strip_quotes:
                return return_dt.lstrip().rstrip()
            else:
                return return_dt.replace('"""', "").rstrip().lstrip() if return_dt.find('"""') != -1 else return_dt.replace("'''", "").rstrip().lstrip()
        except UnboundLocalError:
            return ""
    
    def _outer_indent(self, code):
        spaces_arr = []
        matches = re.finditer(regex, code)
        for _, match in enumerate(matches, start=1):
            spaces_arr.append(len(match.group(0)))
        return min(spaces_arr) if spaces_arr else 4

    
    def get_all_class_method_names(self) -> List:
        """
        Gets all the method names from a file. 

        A method is a function defined inside a class
        """
        captures = self._run_query_and_get_captures('all_class_methods', self.root_node)
        ret_struct = {}
        current_key = ""
        for tpl in captures:
            if tpl[1] == "class.name":
                current_key = match_from_span(tpl[0], self.splitted_code)
                ret_struct[current_key] = []
                continue
            else:
                ret_struct[current_key].append(match_from_span(tpl[0], self.splitted_code))
        return ret_struct

    def get_all_function_names(self) -> List:
        """
        Gets all function names from a file.

        It excludes all the methods, i.e. functions defined inside a class
        """
        captures = self._run_query_and_get_captures('all_function_names', self.root_node)
        all_funcs = set([match_from_span(n[0], self.splitted_code) for n in captures])

        methods = self.get_all_class_method_names()
        all_methods = set([method_name  for key, value in methods.items() for method_name in value])

        return list(all_funcs - all_methods)
    
    def get_all_function_docstrings(self, strip_quotes: bool=False) -> Dict:
        """
        Returns a dict where function names are the key and the docstrings are the values

        Excludes any methods, i.e., functions defined inside a class.

        Optional argugmet "strip_quotes" gives the choice whether the docstring returned 
        will be strippted out of tripple quotes or not. Default: False
        """
        function_names = self.get_all_function_names()
        captures = self._run_query_and_get_captures('all_function_doctrings', self.root_node)
        ret_struct = {}
        for i in range(0, len(captures), 2):
            func_name = match_from_span(captures[i][0], self.splitted_code)
            if func_name in function_names:
                ret_struct[func_name] = self._strip_py_doc_string(match_from_span(
                                                                  captures[i+1][0], self.splitted_code
                                                                  ), strip_quotes)
        return ret_struct
    
    def get_all_method_docstrings(self, strip_quotes: bool=False) -> Dict:
        """
        Returns a dict where method names are the key and the docstrings are the values

        Excludes any functions, i.e., functions defined outside a class.

        Optional argugmet "strip_quotes" gives the choice whether the docstring returned 
        will be strippted out of tripple quotes or not. Default: False
        """
        captures = self._run_query_and_get_captures('all_class_method_docstrings', self.root_node)
        ret_struct = {}
        current_class = ""
        current_method = ""
        for tpl in captures:
            if tpl[1] == "class.name":
                current_class = match_from_span(tpl[0], self.splitted_code)
                ret_struct[current_class] = {}
                continue
            elif tpl[1] == "method.name":
                current_method = match_from_span(tpl[0], self.splitted_code)
                ret_struct[current_class][current_method] = ""
                continue
            elif tpl[1] == "method.docstr":
                ret_struct[current_class][current_method] = self._strip_py_doc_string(match_from_span(
                                                                                      tpl[0], self.splitted_code
                                                                                      ), strip_quotes)
        return ret_struct
    
    def get_all_function_bodies(self, strip_docstr: bool=False) -> Dict:
        """
        Returns a dict where function names are the key and the whole function code are the values

        Excludes any methods, i.e., functions defined inside a class.

        Optional argugmet "strip_docstr" gives the choice whether the docstring should
        be returned as a part of the function body or separately.
        """
        function_names = self.get_all_function_names()
        func_and_params = self.function_names_with_params()
        func_and_docstr = self.get_all_function_docstrings()
        
        captures = self._run_query_and_get_captures('all_function_bodies', self.root_node)
        
        ret_struct = {}
        pp = {}
        for i in range(0, len(captures), 2):
            func_name = match_from_span(captures[i][0], self.splitted_code)
            if func_name in function_names:
                pp[func_name] = match_from_span(captures[i+1][0], self.splitted_code)
        
        if strip_docstr:
            for k, v in pp.items():
                if func_and_docstr.get(k) is not None and func_and_docstr.get(k) is not '':
                    code = v.replace(func_and_docstr[k], "")
                    outer_indent = self._outer_indent(code)
                    spaces = " ".join([''] * (outer_indent + 1))
                    if code.startswith("\n"):
                        ret_struct[k] = (f"def {k}{func_and_params[k]}:{code}", func_and_docstr[k])
                    else:
                        ret_struct[k] = (f"def {k}{func_and_params[k]}:\n{spaces}{code}", func_and_docstr[k])
                else:
                    outer_indent = self._outer_indent(v)
                    spaces = " ".join([''] * (outer_indent + 1))
                    ret_struct[k] = (f"def {k}{func_and_params[k]}:\n{spaces}{v}", "")
        else:
            for k, v in pp.items():
                outer_indent = self._outer_indent(v)
                spaces = " ".join([''] * (outer_indent + 1))
                ret_struct[k] = f"def {k}{func_and_params[k]}:\n{spaces}{v}"
        return ret_struct
    
    def function_names_with_params(self, split_params_in_list: bool=False):
        """
        Returns a dictionary with all the function names and their params

        Optional argument split_params_in_list lets the params to be returned as a list
        """
        function_names = self.get_all_function_names()
        captures = self._run_query_and_get_captures('all_function_names_and_params', self.root_node)
        ret_struct = {}

        for i in range(0, len(captures), 2):
            func_name = match_from_span(captures[i][0], self.splitted_code)
            if func_name in function_names:
                params = match_from_span(captures[i+1][0], self.splitted_code)
                if split_params_in_list:
                    if params.startswith("(") and params.endswith(")"):
                        params = params[1:-1]
                        params = params.split(",")
                ret_struct[func_name] = params
        
        return ret_struct
    
    def get_all_class_names(self) -> List[str]:
        """
        Returns a list of all class names present in a file
        """
        captures = self._run_query_and_get_captures('all_class_names', self.root_node)
        return [match_from_span(t[0], self.splitted_code) for t in captures]
    
    def get_all_class_docstrings(self, strip_quotes: bool=False) -> Dict[str, str]:
        """
        Returns the docstring of all classes

        Optional argument strip_quotes strips the tripple quotes from both sides
        """
        captures = self._run_query_and_get_captures("all_class_docstrings", self.root_node)

        ret_struct = {}
        for i in range(0, len(captures), 2):
            class_name = match_from_span(captures[i][0], self.splitted_code)
            class_body_node = captures[i+1][0]
            if self._has_children(class_body_node):
                cursor = class_body_node.walk()
                if cursor.goto_first_child() and cursor.node.type == "expression_statement":
                    if cursor.goto_first_child() and cursor.node.type == "string":
                        class_docstr = match_from_span(cursor.node, self.splitted_code)
                        if strip_quotes:
                            class_docstr = self._strip_py_doc_string(class_docstr, strip_quotes)
                        ret_struct[class_name] = class_docstr
        return ret_struct